/* Copyright (C) 2006 Anton Fedorov <datacompboy@call2ru.com>

   nss_vserver: provide auth on host according to passwd & shadow files
		in vservers.
   You can find info about vserver at http://linux-vserver.org/
   See README for details
   
   Feel free to use code as you want
 */


#include <nss.h>
#include <pthread.h>
#include <glob.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "commdef.h"

typedef struct ENT_CONTEXT {
  glob_t globbuf;
  int globuse;
  int file;
  char * vserver;
  int vserverlen;
  DBOPS_T lastop;
  char * buf;
  int scantry;
  size_t bufsize;
  size_t buflen;
  size_t bufoff;
} ENT_CONTEXT_T;

static void context_init(ENT_CONTEXT_T * context)
{
  context->globbuf.gl_offs = 0;
  context->globuse = 0;
  context->file    = 0;
  context->vserver = NULL;
  context->vserverlen = 0;
  context->lastop  = OP_NOOP;
  context->buf	   = NULL;
  context->bufsize = 0;
  context->buflen  = 0;
  context->bufoff  = 0;
  context->scantry = 0;
}

static void context_free(ENT_CONTEXT_T * context)
{
  if (0!=context->globuse) {
	globfree(&context->globbuf);
  }
  if (NULL!=context->buf) {
	free(context->buf);
  }
  if (0!=context->file) {
	close(context->file);
  }
  context_init(context);
}

#include "commctx.h"

/* Read to buffer from its current offset up to its size */
/* If offset is zero, and file closed or at end of file -- */
/* Open next, if next present */
static int vs_readbuf(ENT_CONTEXT_T * context)
{
  while (1) {
	if (context->buflen == context->bufsize) {
	  /* Buffer already full -- return error */
	  return -1;
	}
	if (context->file != 0) {
	  /* File already open, try to read from */
	  int rr = 0;
	  while (context->buflen < context->bufsize) {
		int r = read(context->file, &context->buf[context->buflen],
					 context->bufsize - context->buflen);
		if (r<=0) break; /* Break on EOF/error */
		rr+=r; /* otherwize count readed and andvance buffer length */
		context->buflen += r;
	  }
	  /* Now check, is we read something or not. */
	  if (rr>0) {
		/* We read -- return count of read */
		return rr;
	  } else {
		/* We read nothing (error in file read, or file at end) */
		if (context->buflen > 0) {
		  /* Buffer was not empty -- we can't go to next file. return EOF */
		  return 0;
		} else {
		  /* We need go to next file */
		  close(context->file);
		  context->file = 0;
		  context->globuse++;
		}
	  }
	}
	/* Our file is closed. We need to open new one (already pointed) */
	if (context->globuse >= context->globbuf.gl_pathc) {
	  /* No more files available */
	  return -2;
	}
	context->file = open(context->globbuf.gl_pathv[context->globuse], O_RDONLY);
	/* On error, go to next file */
	if (context->file < 0) {
	  context->file = 0;
	  context->vserver = 0;
	  context->vserverlen = 0;
	} else {
	  /* otherwise, need to find vserver part in path */
	  char * vse = context->vserver
				 = &context->globbuf.gl_pathv[context->globuse][strlenETCVSERVER];
	  while (*vse != 0 && *vse != '/') vse++;
	  context->vserverlen = vse-context->vserver;
	}
  }
}

/* Check is complete line at buffer present */
/* If not, try to discard beginning of buffer */
/* If file ends, try to open next file, and find its vserver name */
static int vs_readline(ENT_CONTEXT_T * context)
{
  if (0 == context->bufsize || !context->buf) {
	/* I don't think that normal passwd entries will be more than 4K */
	context->bufsize = 4096;
	context->buf = (char*)malloc(context->bufsize*sizeof(char));
	context->bufoff = 0;
	context->buflen = 0;
  }
  while(1) {
	int off = context->bufoff;
	int lend = 0;
	/* Skip emptiness from current position */
	while (off < context->buflen && (
			context->buf[off] == ' ' ||
			context->buf[off] == '\t' ||
			context->buf[off] == '\r' ||
			context->buf[off] == '\n')
		  ) off++;
	if (off == context->buflen) {
	  /* If our buffer is ended, we need to load new buffer */
	  /* completely and restart scan */
	  context->buflen = 0;
	  context->bufoff = 0;
	  if (vs_readbuf(context)<=0) {
		/* EOF, and no more files */
		return 0;
	  } else {
		continue;
	  }
	}
	/* we have reached start of new line, find for its end */
	lend = off;
	while (lend < context->buflen && context->buf[lend] != '\n') lend++;
	/* check, is we find end of buffer, not end of line */
	if (lend == context->buflen) {
	  /* if so, try to read all line to buffer */
	  memmove(context->buf, &context->buf[off], lend-off+1);
	  lend -= off; off = 0;
	  context->buflen = lend;
	  context->bufoff = off;
	  if (vs_readbuf(context)<=0) {
		/* EOF, or buffer full -- suggest that line is that we have now */
	  } else {
		/* we have readed something, try to find eol here */
		while (lend < context->buflen && context->buf[lend] != '\n') lend++;
		/* now no difference in end of buf, or end of line -- we unable */
		/* to read more than buf space have */
	  }
	}
	if (context->buf[off]=='#') {
	  /* if current line is comment -- skip it */
	  context->bufoff=lend;
	  continue;
	  /* BUG! If passwd have comment line with length more than 4K, then */
	  /* Skipped only 4K, rest is threaded as normal passwd line */
	  /* But this is not normal */
	}
	/* If we here -- we have line in buffer, return its length */
	context->bufoff = off;
	return lend-off;
  }
}

NSS_STATUS CONCAT3(_nss_vserver_set,NSS_DB,ent) (void)
{
  ENT_CONTEXT_T * context = get_context();
  if (context->lastop!=OP_NOOP) {
	context_free(context);
  }
  context->lastop=OP_SETENT;
  glob(ETCVSERVER "*/vdir/etc/" NSS_FILES, GLOB_NOSORT, NULL, &context->globbuf);
  if (context->globbuf.gl_pathc > 0) {
	return NSS_STATUS_SUCCESS;
  } else {
	return NSS_STATUS_UNAVAIL;
  }
}

NSS_STATUS CONCAT3(_nss_vserver_end,NSS_DB,ent) (void)
{
  ENT_CONTEXT_T * context = get_context();
  context_free(context);
  return NSS_STATUS_SUCCESS;
}

NSS_STATUS CONCAT3(_nss_vserver_get,NSS_DB,ent_r) (struct NSS_RES *result,
								 char *buffer, size_t buflen, int *errnop)
{
  ENT_CONTEXT_T * context = get_context();
  if (!(context->lastop == OP_SETENT || context->lastop==OP_GETENT)) {
	*errnop=EPROTO;
	return NSS_STATUS_UNAVAIL;
  }
  context->lastop = OP_GETENT;
  while (1) {
	int len = vs_readline(context); /* len is including newline, if exist */
	int parse = 0;
	/* We need to get at least one full line from file at one vserver */
	/* vs_readline skip comment and empty lines automatically */
	if (len <= 0) {
	  *errnop=ENOENT;
	  return NSS_STATUS_NOTFOUND;
	}
	/* We call local parser, that also should check correctness of line */
	parse = parse_line(result, buffer, buflen, &context->buf[context->bufoff], len, context->vserver, context->vserverlen, context->scantry++);
	if (2 == parse) {
	  /* That is correct line, but we have some more variants of parse */
	  *errnop=0;
	  return NSS_STATUS_SUCCESS;
	} else
	if (1 == parse) {
	  /* If line is correct entry, return success and advance to next entry */
	  context->bufoff += len;
	  context->scantry=0;
	  *errnop=0;
	  return NSS_STATUS_SUCCESS;
	} else
	if (0 == parse) {
	  /* Line correct, but there no space to store it in buffer, request more */
	  *errnop=ERANGE;
	  return NSS_STATUS_TRYAGAIN;
	} else {
	  /* Incorrect line -- skip it and try next */
	  context->bufoff += len;
	}
  }
}

NSS_STATUS CONCAT3(_nss_vserver_get,NSS_DB,nam_r) (const char *name,
								 struct NSS_RES * result,
								 char *buffer, size_t buflen, int *errnop)
{
  ENT_CONTEXT_T * context = get_context();
  char * vserver; int vserverlen;
  char * uname; int unamelen;
  if (context->lastop!=OP_NOOP) {
	context_free(context);
  }
  context->lastop=OP_GETNAM;
  /* for getnam we should parse name, split it to name and vserver parts */
  /* then search only that vserver's file */
  /* We will use buffer for splitting name */
  if (strlen(name)+1 > buflen) {
	  *errnop=ERANGE;
	  return NSS_STATUS_TRYAGAIN;
  }
  /* parse_user: split given name to username (should be put first in buffer) */
  /* and vserver (put into buffer after username) */
  if (parse_user((char*)name, &vserver, &vserverlen, &uname, &unamelen)) {
	/* Non-vserver username: we doesn't handle them */
	*errnop=ENOENT;
	return NSS_STATUS_NOTFOUND;
  }
  /* Here safe: we already checked size of buffer */
  strcpy(buffer, name);
  /* We should try to find file for only current vserver */
  {
	char globpath[16384] = ETCVSERVER;
	char * buf = globpath+strlenETCVSERVER;
	if (strlenETCVSERVER+sizeof("/vdir/etc/" NSS_FILES)+vserverlen >= sizeof(globpath)) {
	  *errnop = ENOMEM;
	  return NSS_STATUS_UNAVAIL;
	}
	memmove(buf, vserver, vserverlen); buf+=vserverlen;
	memmove(buf, "/vdir/etc/" NSS_FILES, sizeof("/vdir/etc/" NSS_FILES));
	  /* Zero included in last move operation */
	glob(globpath, GLOB_NOSORT, NULL, &context->globbuf);
	if (context->globbuf.gl_pathc <= 0) {
	  *errnop = ENOENT;
	  return NSS_STATUS_UNAVAIL;
	}
  }
  /* Now, loop over file and see */
  while (1) {
	int len = vs_readline(context); /* len is including newline, if exist */
	int parse = 0;
	if (len <= 0) {
	  *errnop=ENOENT;
	  return NSS_STATUS_NOTFOUND;
	}
	/* We call local parser, that also should check correctness of line */
	/* parse_line_cmp differs from parse_line by comparing uname with	*/
	/* key passed as last arg */
	parse = parse_line_cmp(result, buffer, buflen, &context->buf[context->bufoff], len, context->vserver, context->vserverlen, uname, unamelen);
	if (1 == parse) {
	  /* If line is correct entry, return success and advance to next entry */
	  context->bufoff += len;
	  *errnop=0;
	  return NSS_STATUS_SUCCESS;
	} else
	if (0 == parse) {
	  /* Line correct, but there no space to store it in buffer, request more */
	  *errnop=ERANGE;
	  return NSS_STATUS_TRYAGAIN;
	} else {
	  /* Incorrect line -- skip it and try next */
	  context->bufoff += len;
	}
  }
}
